#!/usr/local/bin/python
#
# $Id: //projects/csbuddy/csbuddy.py#18 $ $Date: 2003/04/02 $

"""
Monitoring system for Counter-Strike logs.
"""

__program__ = 'csbuddy'
__version__ = '1.2.5'
__url__ = 'http://www.alcyone.com/pyos/csbuddy/'
__author__ = 'Erik Max Francis <max@alcyone.com>'
__copyright__ = 'Copyright (C) 2002 Erik Max Francis'
__license__ = 'GPL'


import getopt
import os
import cPickle as pickle
import re
import sys
import time

import cslog


try:
    BASE_DIRECTORY = os.environ['HOME'] + '/.csbuddy'
except KeyError:
    BASE_DIRECTORY = './.csbuddy'

FLOODING_THRESHHOLD = 5
INCIDENTS_THRESHHOLD = 10
DETECT_WRAP = 1


class MetaFile:
    def __init__(self):
        self.streamsMap = {}
        self.streamsMap['-'] = sys.__stdout__
        self.streamsMap['default'] = open(subfile('log'), 'a')
        self.streamsMap['chat'] = open(subfile('chat'), 'a')
        self.buffer = ''
            
    def process(self):
        """Process the internal buffer."""
        loc = self.buffer.find('\n')
        if loc >= 0:
            line, self.buffer = self.buffer[:loc], self.buffer[loc + 1:]
            stream, line = line.split('#', 1)
            print >> self.streamsMap[stream], line

    def write(self, data):
        self.buffer += data
        self.process()

    def writelines(self, lines):
        for line in lines:
            self.write(line)


class Player:
    def __init__(self, won):
        self.won = won
        self.names = {}
        self.ips = {}
        self.firstConnect = None
        self.lastConnect = self.firstConnect
        self.lastDisconnect = 0
        self.lastName = ''
        self.totalTime = 0.0
        self.totalRounds = 0
        self.kills = 0
        self.deaths = 0
        self.totalIncidents = 0
        self.sayHistory = None, 0

    def ID(self):
        """Get an ID for the player, consisting of the WONID and the name."""
        return '%d/%s' % (self.won, self.name())

    def name(self):
        """The player's current name."""
        return self.lastName

    def mostUsedName(self):
        """Find the player's most frequently used name."""
        items = self.names.items()
        items.sort(lambda x, y: cmp(y[1], x[1]))
        if items:
            return items[0][0]
        else:
            return '?'

    def firstSeen(self):
        """When was the player first seen."""
        return self.firstConnect
    
    def lastSeen(self):
        """When was the player last seen."""
        return max(self.lastConnect, self.lastDisconnect)

    def ratio(self):
        """The player's kill/death ratio."""
        if self.deaths != 0:
            return float(self.kills)/float(self.deaths)
        return 0.0

    def registerName(self, name):
        """Register a new name for this player."""
        if name == 'ADMIN' or name == 'admin' or name == '?':
            return
        self.lastName = name
        if self.names.has_key(name):
            self.names[name] += 1
        else:
            self.names[name] = 1

    def connect(self, when, ip):
        """Register a new connect."""
        if self.ips.has_key(ip):
            self.ips[ip] += 1
        else:
            self.ips[ip] = 1
        if self.firstConnect is None:
            self.firstConnect = when
        self.lastConnect = when

    def disconnect(self, when):
        """Register a disconnect."""
        self.lastDisconnect = when
        if self.lastConnect is not None:
            self.totalTime += when - self.lastConnect

    def say(self, message):
        """Do a check for flooding, return None if nothing to report, else
        a 2-tuple of (what, times)."""
        what, times = self.sayHistory
        if what == message:
            self.sayHistory = what, times + 1
        else:
            self.sayHistory = message, 1
            if times >= FLOODING_THRESHHOLD:
                return what, times
        return None

    def show(self, verbose=1):
        """Show info on this player."""
        print "Player:", self.ID()
        names = self.names.items()
        names.sort(lambda x, y: cmp(y[1], x[1]))
        print "Names: %s" % ', '.join(["%s (%d)" % (x[0], x[1]) for x in names])
        ips = self.ips.items()
        ips.sort(lambda x, y: cmp(y[1], x[1]))
        print "IP addresses: %s" % ', '.join(["%s (%d)" % (x[0], x[1]) for x in ips])
        if verbose:
            firstSeen = self.firstSeen()
            if firstSeen is None:
                firstSeenStr = 'never'
            else:
                firstSeenStr = time.ctime(firstSeen)
            print "First seen: %s" % firstSeenStr
            print "Last seen: %s" % time.ctime(self.lastSeen())
            print "Rounds: %d" % self.totalRounds
            print "Kills/deaths: %d/%d (%.2f)" % \
                  (self.kills, self.deaths, self.ratio())
            print "Total time: %.2f hours" % (self.totalTime/3600.0,)
            print "Total incidents: %d" % self.totalIncidents
            if self.totalRounds:
                print "Total incidents per round: %.2f" % (float(self.totalIncidents)/self.totalRounds,)
            if self.totalTime:
                print "Total incidents per hour: %.2f" % (float(self.totalIncidents)/(self.totalTime/3600.0),)
            print "Last said: \"%s\"" % self.sayHistory[0]


class Buddy(cslog.NullParser):
    ADMIN_MENTIONED_RE = re.compile(r'\b(owners?|admins?|pro(d(uct)?)?)\b', re.IGNORECASE)
    PROSCRIBED_KEYWORD_RE = re.compile(r'((hl|cs)guard|cheat|tk|flood|kick|ban)', re.IGNORECASE)
    
    def __init__(self):
        cslog.NullParser.__init__(self)
        self.current = None # the last log file seen
        self.players = {} # map of wons to players
        self.totalMaps = 0l # total maps played
        self.totalRounds = 0l # total rounds played
        self.reset()

    # Scanner methods.

    def isValidFilename(self, filename):
        """Is filename a valid log file?"""
        return len(filename) == 12 and \
               filename.startswith('L') and filename.endswith('.log')

    def extractName(self, filename):
        """Extract the name from a valid filename."""
        return filename[1:-4]

    def path(self, directory, name):
        """Extract the name from a valid filename."""
        return '%s/L%s.log' % (directory, name)

    def collect(self, directory):
        """Collect files."""
        filenames = os.listdir(directory)
        validFilenames = filter(self.isValidFilename, filenames)
        culledNames = map(self.extractName, validFilenames)
        culledNames.sort()
        return culledNames

    def scan(self, directory, since=None):
        """Get the names since the specified log."""
        if since is None:
            since = self.current
        names = self.collect(directory)
        if names and names[0].startswith('01') and names[-1].startswith('12') \
               and DETECT_WRAP and since != '':
            # If there are logs from both January and December, then we're
            # in the special rollover case where the logs span a year
            # boundary.  Presume that there is at least one full month that's
            # missing, and use that as the place to split.
            months = {}
            for m in range(1, 12 + 1):
                months[m] = None
            # Build a list of the months, and the first item in the list
            # that starts them, or None.
            for i in range(len(names)):
                name = names[i]
                month = int(name[:2])
                if months[month] is None:
                    months[month] = i
            # Start backward from the end, and find the first month which had
            # no log files.  That's the break between the old and new.
            for j in range(12, 0, -1):
                if months[j] is None:
                    break
            else:
                raise ValueError, "clean up log files; cannot find boundary"
            # Now rearrange the names so that they're in the proper ordering,
            # crossing the year boundary (e.g., ..., 11, 12, 01, 02, ...).
            assert j != 12
            first = months[j + 1]
            names = names[first:] + names[:first]
        # Skip the latest file, since it's probably still in process.
        names = names[:-1]
        if since is not None:
            # A since file of '' means everything (force mode).
            if since == '':
                return names
            # If a since file has been specified, try to find it.
            newNames = []
            try:
                # Start by just calling the .index method; if the name itself
                # is there, great.
                index = names.index(since)
                newNames = names[index + 1:]
            except ValueError:
                # Otherwise, it's not in the list, so we need to go by
                # sort order.
                for name in names:
                    if name > since:
                        newNames.append(name)
            return newNames
        else:
            # If nothing was specified, just use everything.
            return names

    def hasNew(self, directory, force=1):
        """Are there new files available?"""
        if force:
            return 1
        else:
            return len(self.scan(directory))

    def sync(self, directory, force=0):
        """Sync up to the latest log file."""
        if force:
            since = ''
        else:
            since = self.current
        names = self.scan(directory, since)
        if names:
            for name in names:
                file = open(self.path(directory, name), 'r')
                self.current = name
                self.go(file)
                self.flush()

    # State methods.

    def hasWon(self, won):
        """Has the given WON been seen before?"""
        return self.players.has_key(won)

    def player(self, won):
        """Get the player object with the specified WONID, or create a
        new one."""
        if not self.hasWon(won):
            self.players[won] = Player(won)
        return self.players[won]

    def findPlayerByWon(self, won):
        """Find the unique player specified by the WONID, or None."""
        if self.hasWon(won):
            return self.player(won)
        else:
            return None

    def findPlayersByName(self, name):
        """Get a list of all players that have used that name."""
        players = []
        for player in self.players.values():
            if name in player.names.keys():
                players.append(player)
        return players

    def findPlayersByPattern(self, pattern):
        """Get a list of players that have used a similar name."""
        pattern = pattern.lower()
        players = []
        for player in self.players.values():
            for name in player.names.keys():
                if name.lower().find(pattern) >= 0:
                    players.append(player)
                    break
        return players

    def findPlayersByIP(self, ip):
        """Get a list of all players that have used that IP address."""
        players = []
        for player in self.players.values():
            if ip in player.ips.keys():
                players.append(player)
        return players

    def findPlayersByIPPrefix(self, ipPrefix):
        """Get a list of all players that have used that IP address."""
        players = []
        for player in self.players.values():
            for ip in player.ips.keys():
                if ip.startswith(ipPrefix):
                    players.append(player)
                    break
        return players

    def findPlayerByName(self, name):
        """Find the player by name, using a heuristic for determining what
        the most likely name is."""
        if name == 'ADMIN' or name == 'Admin' or name == '?':
            return None
        # If someone with this name is still on, then use it.
        for won in self.currentWons:
            player = self.player(won)
            if player.lastName == name:
                return player
        # Otherwise build a list of everyone who has used that name,
        # and pick the one that has used it the most.
        possibles = self.findPlayersByName(name)
        if possibles:
            possibles.sort(lambda x, y, N=name: cmp(y.names[N], x.names[N]))
            return possibles[0]
        else:
            return None

    def findPlayerByIP(self, ip):
        """Find the player by IP address, using the most frequently used
        IP address."""
        if ip == '127.0.0.1':
            return None
        possibles = self.findPlayersByIP(ip)
        if possibles:
            possibles.sort(lambda x, y, I=ip: cmp(y.ips[I], x.ips[I]))
            return possibles[0]
        else:
            return None

    def reset(self):
        """Reset the parser."""
        self.finished = 1
        self.map = None
        self.currentWons = []
        self.rounds = 0
        self.incidents = {}
        self.deferredLocals = {}
        
    def flush(self):
        """We're finished with a file, finalize up the parser."""
        if not self.finished:
            self.report("POSSIBLE CRASH")
        self.rounds = 999
        for won, count in self.incidents.items():
            if count >= INCIDENTS_THRESHHOLD:
                player = self.player(won)
                self.report("FRIENDLY FIRE: {%s}: %d incidents" % (player.ID(), count))
        self.reset()

    # Overridden Parser methods.

    def friendlyFire(self, attacker, count=1):
        if self.incidents.has_key(attacker.won):
            self.incidents[attacker.won] += count
        else:
            self.incidents[attacker.won] = count
        attacker.totalIncidents += count

    def eventLogStart(self):
        self.finished = 0

    def eventLogEnd(self):
        self.finished = 1

    def eventStartedMap(self, map, crc):
        self.map = map
        self.totalMaps += 1
        self.report("MAP START: %s" % map)
        self.report("MAP START: %s" % map, stream='chat')
    
    def eventConnect(self, who, address):
        name, local, won, team = who
        if won:
            if won not in self.currentWons:
                self.currentWons.append(won)
            player = self.player(won)
            first = player.firstSeen() is None
            player.connect(self.now, address[0])
            player.registerName(name)
            if first:
                self.report("{%s} connected (new)" % player.ID(), stream='chat')
            else:
                self.report("{%s} connected" % player.ID(), stream='chat')
        else:
            self.deferredLocals[local] = address

    def eventUserValidated(self, who):
        name, local, won, team = who
        if self.deferredLocals.has_key(local):
            address = self.deferredLocals[local]
            del self.deferredLocals[local]
        else:
            address = '?'
        player = self.player(won)
        player.connect(self.now, address[0])
        player.registerName(name)

    def eventDisconnect(self, who):
        name, local, won, team = who
        if won in self.currentWons:
            self.currentWons.remove(won)
        player = self.player(won)
        player.disconnect(self.now)
        self.report("{%s} disconnected" % player.ID(), stream='chat')

    def eventNameChange(self, who, newName):
        name, local, won, team = who
        player = self.player(won)
        self.report("{%s} changed name to \"%s\"" % (player.ID(), newName), \
                    stream='chat')
        player.registerName(newName)

    def eventAttack(self, attacker, defender, weapon, damage, left):
        attackerName, attackerLocal, attackerWon, attackerTeam = attacker
        defenderName, defenderLocal, defenderWon, defenderTeam = defender
        attacker = self.player(attackerWon)
        defender = self.player(defenderWon)
        if attackerTeam == defenderTeam:
            self.friendlyFire(attacker)
        
    def eventKill(self, attacker, defender, weapon):
        attackerName, attackerLocal, attackerWon, attackerTeam = attacker
        defenderName, defenderLocal, defenderWon, defenderTeam = defender
        attacker = self.player(attackerWon)
        defender = self.player(defenderWon)
        if attackerWon != defenderWon:
            ### For now, just skip it if the WONIDs match.
            attacker.kills += 1
            defender.deaths += 1
        if attackerTeam == defenderTeam:
            self.friendlyFire(attacker, 4) # 4 more times for a total of 5

    def eventSuicide(self, who, weapon, wasWorld):
        name, local, won, team = who
        player = self.player(won)
        player.deaths += 1

    def eventSay(self, who, message):
        name, local, won, team = who
        player = self.player(won)
        result = player.say(message)
        if result is not None:
            what, times = result
            self.report("FLOODING: {%s}: \"%s\" (%d times)" % \
                        (player.ID(), what, times))
        if self.ADMIN_MENTIONED_RE.search(message):
            self.report("ADMIN MENTIONED: {%s}: \"%s\"" % (player.ID(), message))
        elif self.PROSCRIBED_KEYWORD_RE.search(message):
            self.report("PROSCRIBED KEYWORD: {%s}: \"%s\"" % \
                        (player.ID(), message))
        self.report("{%s} (%s): \"%s\"" % (player.ID(), team, message), \
                    stream='chat')

    def eventSayTeam(self, who, message):
        name, local, won, team = who
        player = self.player(won)
        result = player.say(message)
        if result is not None:
            what, times = result
            self.report("FLOODING: {%s}: \"%s\" (%d times) [%s]" % \
                        (player.ID(), what, times, team))
        if self.ADMIN_MENTIONED_RE.search(message):
            self.report("ADMIN MENTIONED: {%s}: \"%s\" [%s]" % (player.ID(), message, team))
        elif self.PROSCRIBED_KEYWORD_RE.search(message):
            self.report("PROSCRIBED KEYWORD: {%s}: \"%s\" [%s]" % \
                        (player.ID(), message, team))
        self.report("{%s} [%s]: \"%s\"" % (player.ID(), team, message), \
                    stream='chat')

    def triggerRoundStart(self):
        self.totalRounds += 1
        self.rounds += 1
        for won in self.currentWons:
            player = self.player(won)
            player.totalRounds += 1

    def eventSecureKick(self, who):
        name, local, won, team = who
        player = self.player(won)
        self.report("SECURE KICK: {%s}" % player.ID())

    def adminCommand(self, name, verb, noun):
        if name != 'chat' and not verb.find('vote') >= 0:
            player = self.findPlayerByName(name)
            if player is None:
                playerID = '?'
            else:
                playerID = player.ID()
            verb = verb.lower()
            if verb in ('admin_say', 'admin_ssay', 'admin_csay', 'admin_tsay', 'admin_psay'):
                major = "ADMIN SAY"
                self.report("{%s} <%s>: \"%s\"" % (playerID, verb, noun), \
                            stream='chat')
            elif verb in ('admin_llama', 'admin_gag', 'admin_slap', 'admin_slay', 'admin_slayteam', 'admin_kick', 'admin_ban'):
                major = "ADMIN PUNISH"
            elif verb == 'admin_map':
                major = "ADMIN MAP CHANGE"
            else:
                if player is None:
                    major = None
                else:
                    major = "ADMIN COMMAND"
            if major:
                self.report("%s: {%s}: %s %s" % (major, playerID, verb, noun))

    def adminSay(self, type, text):
        if type == 'chat' and text.find(':') >= 0:
            name, message = text.split(':', 1)
            message = message.strip()
            player = self.findPlayerByName(name)
            if player is None:
                playerID = '?'
            else:
                playerID = player.ID()
            self.report("ADMIN ANNOTATION: {%s}: %s" % (playerID, message))
            self.report("{%s} <admin_chat>: \"%s\"" % (playerID, message), \
                        stream='chat')

    def hlguardFound(self, who, what):
        name, local, won, team = who
        player = self.player(won)
        self.report("HLGUARD FOUND: {%s}: %s" % (player.ID(), what))

    def hlguardBanned(self, who):
        name, local, won, team = who
        player = self.player(won)
        self.report("HLGUARD BAN: {%s}" % player.ID())

    # Reporter methods.

    def report(self, message, when=None, stream='default'):
        if when is None:
            when = self.now
        specifier = "%s.%03d" % (self.current, self.rounds)
        stamp = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(when))
        print "%s#%s %s | %s" % (stream, stamp, specifier, message)

    def summary(self):
        """Return a tuple of: current log file, total unique WONIDs, total
        maps, and total rounds."""
        return self.current, len(self.players), \
               self.totalMaps, self.totalRounds


def subfile(name):
    """Get subfile filename."""
    return BASE_DIRECTORY + '/csbuddy.' + name


def summary():
    """Load the summary."""
    try:
        file = open(subfile('state'), 'r')
        unpickler = pickle.Unpickler(file)
        summary = unpickler.load()
        return summary
    except IOError:
        return None, 0, 0L, 0L

def load():
    """Load a full pickled state."""
    try:
        file = open(subfile('state'), 'r')
        unpickler = pickle.Unpickler(file)
        unpickler.load() # dump the summary
        buddy = unpickler.load()
        return buddy
    except IOError:
        print >> sys.stderr, "%s: no state file, starting fresh" % sys.argv[0]
        return Buddy()

def save(buddy):
    """Save the current state."""
    newStateFilename = subfile('state') + '.new'
    file = open(newStateFilename, 'w')
    unpickler = pickle.Pickler(file, 1)
    unpickler.dump(buddy.summary())
    unpickler.dump(buddy)
    file.close()
    os.rename(newStateFilename, subfile('state'))

def usage():
    print >> sys.stderr, """\
Usage: %s [options] <mode> <args...>

Options:
  -V --version      Print version and exit
  -h --help         Print help and exit
  -b --brief        Brief output
  -f --force        Force full scan
  -d --dir=<dir>    Specify a new base directory

Mode options:
  -s --summary      Print summary and exit
  -S --sync         Sync database against log file directories
  -w --won          Find and print info on player with WONIDs
  -n --name         Find player who most used specified names
  -N --names        Find all players who have ever used specified names
  -P --patterns     Find all players containing substrings (case insensitive)
  -i --ip           Find player who mostly used specified IP addresses
  -I --ips          Find all players who have ever used specified IP addresses
  -J --prefixes     Find all players who have ever used specified IP prefix
""" % sys.argv[0],

def main():
    global BASE_DIRECTORY
    verbose = 1
    force = 0
    mode = 'summary'
    pairs, remainder = getopt.getopt(sys.argv[1:], 'Vhbfd:sSwnNPiIJ', ['version', 'help', 'brief', 'force', 'dir=', 'directory=', 'summary', 'sync', 'won', 'name', 'names', 'patterns', 'ip', 'ips', 'prefixes'])
    for option, argument in pairs:
        if option in ('-V', '--version'):
            sys.stderr.write("%s version %s\n" % (__program__, __version__))
            return
        elif option in ('-h', '--help'):
            usage()
            return
        elif option in ('-b', '--brief'):
            verbose = 0
        elif option in ('-f', '--force'):
            force = 1
        elif option in ('-d', '--dir', '--directory'):
            BASE_DIRECTORY = argument
        elif option in ('-s', '--summary'):
            mode = 'summary'
        elif option in ('-S', '--sync'):
            mode = 'sync'
        elif option in ('-w', '--won'):
            mode = 'won'
        elif option in ('-n', '--name'):
            mode = 'name'
        elif option in ('-N', '--names'):
            mode = 'names'
        elif option in ('-p', '--pattern'):
            mode = 'pattern'
        elif option in ('-P', '--patterns'):
            mode = 'patterns'
        elif option in ('-i', '--ip'):
            mode = 'ip'
        elif option in ('-I', '--ips'):
            mode = 'ips'
        elif option in ('-J', '--ip-prefixes'):
            mode = 'prefixes'
    if mode != 'summary' and not remainder:
        raise ValueError, "%s mode requires parameters" % mode
    if mode == 'summary':
        current, totalWons, totalMaps, totalRounds = summary()
        if current is not None:
            print "Last log file processed: %s" % current
            print "Total unique WONIDs: %d" % totalWons
            print "Total maps: %d" % totalMaps
            print "Total rounds: %d" % totalRounds
        else:
            print >> sys.stderr, "No state file exists!"
    elif mode == 'sync':
        sys.stdout = MetaFile()
        buddy = load()
        for directory in remainder:
            if buddy.hasNew(directory, force):
                buddy.sync(directory, force)
                file = open(subfile('names'), 'w')
                wons = buddy.players.keys()
                wons.sort()
                for won in wons:
                    print >> file, '%d\t%s' % (won, buddy.players[won].name())
                file.close()
        save(buddy)
    elif mode == 'won':
        buddy = load()
        for wonStr in remainder:
            won = int(wonStr)
            player = buddy.findPlayerByWon(won)
            if player:
                player.show(verbose)
                print
            else:
                print >> sys.stderr, "WON %d not found" % won
    elif mode == 'name':
        buddy = load()
        for name in remainder:
            player = buddy.findPlayerByName(name)
            if player:
                player.show(verbose)
                print
            else:
                print >> sys.stderr, "Name %s not found" % name
    elif mode == 'names':
        buddy = load()
        for name in remainder:
            players = buddy.findPlayersByName(name)
            if players:
                for player in players:
                    player.show(verbose)
                    print
            else:
                print >> sys.stderr, "Name %s not found" % name
    elif mode == 'patterns':
        buddy = load()
        for pattern in remainder:
            players = buddy.findPlayersByPattern(pattern)
            if players:
                for player in players:
                    player.show(verbose)
                    print
            else:
                print >> sys.stderr, "Pattern %s not found" % pattern
    elif mode == 'ip':
        buddy = load()
        for ip in remainder:
            player = buddy.findPlayerByIP(ip)
            if player:
                player.show(verbose)
                print
            else:
                print >> sys.stderr, "IP address %s not found" % ip
    elif mode == 'ips':
        buddy = load()
        for ip in remainder:
            players = buddy.findPlayersByIP(ip)
            if players:
                for player in players:
                    player.show(verbose)
                    print
            else:
                print >> sys.stderr, "IP address %s not found" % ip
    elif mode == 'prefixes':
        buddy = load()
        for prefix in remainder:
            if not prefix.endswith('.'):
                prefix = prefix + '.'
            players = buddy.findPlayersByIPPrefix(prefix)
            if players:
                for player in players:
                    player.show(verbose)
                    print
            else:
                print >> sys.stderr, "IP prefix %s not found" % prefix

if __name__ == '__main__': main()
